/*****************************************************************************
 *            Created: 6/05/2012
 *          Copyright: (C) Copyright David Goodwin, 2012
 *            License: GNU General Public License
 *****************************************************************************
 *
 *   This is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ****************************************************************************/

#include "lifewidget.h"
#include <stdlib.h>
#include <QPainter>
#include <QBrush>
#include <QMouseEvent>
#include <QtConcurrentRun>
#include <QMessageBox>
#include <QByteArray>
#include <QtDebug>

#define CELL_SIZE 10


LifeWidget::LifeWidget(QWidget *parent) :
    QWidget(parent)
{
    setMouseTracking(true);
    grid = 0;
    secondary_grid = 0;
    template_grid = 0;

    hover_x = -1;
    hover_y = -1;

    old_hover_x = -1;
    old_hover_y = -1;

    template_mode = false;
    template_width = 0;
    template_height = 0;

    mouse_is_in_widget = false;

    init();
}

void LifeWidget::saveGrid(QString filename) {
    save_grid_v1(grid,grid_width,grid_height,filename);
}

void LifeWidget::loadGrid(QString filename) {
    grid_mutex.lock();
    free_grid(grid,grid_width);
    free_grid(secondary_grid, grid_width);
    grid = 0; secondary_grid = 0;

    grid = load_grid(&grid_width,&grid_height,filename);

    if (grid == NULL) {
        init();
    }

    secondary_grid = duplicate_grid(grid,grid_width,grid_height);
    grid_mutex.unlock();

    redraw_pixmap();
    update();
}

void LifeWidget::placeTemplate(QString filename) {
    if (template_grid) {
        free_grid(template_grid, template_width);
        template_grid = 0;
    }

    template_grid = load_grid(&template_width, &template_height, filename);
    redraw_pixmap(&template_pix, template_grid, template_width,
                  template_height, true);
    template_mode = true;
}

void LifeWidget::clearTemplate() {
    if (template_grid) {
        free_grid(template_grid, template_width);
        template_grid = 0;
    }
    template_mode = false;
    template_width = 0;
    template_height = 0;
    template_pix = QPixmap();
}

void LifeWidget::saveTemplate(QString filename, QString description) {
    uint32_t template_width, template_height;
    grid_t template_grid;

    grid_mutex.lock();

    template_grid = crop_grid(grid, grid_width, grid_height,
                     &template_width, &template_height);
    grid_mutex.unlock();

    save_grid(template_grid, template_width, template_height, filename,
              description);

    free_grid(template_grid, template_width);
    template_grid = 0;
}

void LifeWidget::randomize() {
    grid_mutex.lock();
    for (unsigned int x = 0; x < grid_width; x++) {
        for (unsigned int y = 0; y < grid_height; y++) {
            int random = rand() % 100 + 1;
            if (random < 50)
                grid[x][y] = true;
        }
    }
    grid_mutex.unlock();
    redraw_pixmap();
    update();
}

void LifeWidget::reset() {
    grid_mutex.lock();
    for (unsigned int x = 0; x < grid_width; x++) {
        for (unsigned int y = 0; y < grid_height; y++) {
            grid[x][y] = false;
        }
    }
    grid_mutex.unlock();
    redraw_pixmap();
    update();
}

void LifeWidget::init() {
    grid_width = width() / CELL_SIZE;
    grid_height = height() / CELL_SIZE;

    grid = make_grid(grid_width, grid_height);
    secondary_grid = duplicate_grid(grid,grid_width,grid_height);
}

LifeWidget::~LifeWidget() {
    if (grid)
        free_grid(grid,grid_width);
    if (secondary_grid)
        free_grid(secondary_grid, grid_width);
    if (template_grid)
        free_grid(template_grid, template_width);
    grid = 0; secondary_grid = 0; template_grid = 0;
}

void LifeWidget::resizeEvent(QResizeEvent *) {
    grid_mutex.lock();
    unsigned int new_width = width() / CELL_SIZE;
    unsigned int new_height = height() / CELL_SIZE;

    if (grid == 0) return;

    grid_t new_grid = resize_grid(grid,
                                  grid_width,
                                  grid_height,
                                  new_width,
                                  new_height);

    free_grid(grid, grid_width);
    free_grid(secondary_grid, grid_width);
    grid = 0; secondary_grid = 0;

    grid = new_grid;
    grid_width = new_width;
    grid_height = new_height;

    secondary_grid = duplicate_grid(grid, grid_width, grid_height);
    grid_mutex.unlock();
    update();
}

void LifeWidget::crop() {
    grid_t old_grid = grid;
    uint32_t old_width = grid_width;
    uint32_t old_height=  grid_height;

    grid_mutex.lock();

    grid = crop_grid(old_grid,old_width,old_height,&grid_width,&grid_height);
    free_grid(old_grid, old_width);
    free_grid(secondary_grid, old_width);
    old_grid = 0; secondary_grid = 0;

    secondary_grid = duplicate_grid(grid, grid_width, grid_height);
    grid_mutex.unlock();
    redraw_pixmap(&pix, grid, grid_width, grid_height);
    update();
}

void LifeWidget::redraw_pixmap(QPixmap *pixmap, grid_t grid,
                               uint32_t width, uint32_t height,
                               bool is_template) {
    *pixmap = QPixmap(width * CELL_SIZE, height * CELL_SIZE);

    if (is_template)
        pixmap->fill(Qt::transparent);

    QPainter painter(pixmap);

    if (is_template) {
        painter.setBackgroundMode(Qt::TransparentMode);
        painter.setBrush(QBrush(Qt::red));
        painter.setPen(Qt::white);
    } else {
        painter.setPen(QPen(Qt::white));
        painter.setBrush(QBrush(Qt::white));

        painter.drawRect(0,0,pixmap->width(),pixmap->height());
        painter.setBrush(QBrush(Qt::black));
    }

    for (unsigned int x = 0; x < width; x++) {
        for (unsigned int y = 0; y < height; y++) {
            if (grid[x][y])
                painter.drawRect(x*CELL_SIZE,y*CELL_SIZE,CELL_SIZE,CELL_SIZE);
        }
    }
}

void LifeWidget::redraw_pixmap() {
    redraw_pixmap(&pix, grid, grid_width, grid_height, false);
}


void LifeWidget::paintEvent(QPaintEvent *) {
    QPainter painter(this);

    //pix_mutex.lock();
    painter.drawPixmap(0,0,pix);
    //pix_mutex.unlock();

    if (hover_x >= 0 && hover_y >= 0 && mouse_is_in_widget) {
        if (template_mode) {
            uint32_t x = hover_x, y = hover_y;

            x -= (template_width * CELL_SIZE) / 2;
            y -= (template_height * CELL_SIZE) / 2;

            // Snap to grid
            x -= x % CELL_SIZE;
            y -= y % CELL_SIZE;

            painter.drawPixmap(x, y, template_pix);
        } else {
            painter.setBrush(QBrush(Qt::red));
            painter.drawRect(hover_x,hover_y,CELL_SIZE,CELL_SIZE);
        }
    }
}

void LifeWidget::mouseMoveEvent(QMouseEvent * event) {
    // reduce the coordinates to the next lowest multiple of 20
    int x = event->x();
    x -= x % CELL_SIZE;

    int y = event->y();
    y -= y % CELL_SIZE;

    old_hover_x = hover_x;
    old_hover_y = hover_y;

    hover_x = x;
    hover_y = y;

    mouse_is_in_widget = true;

    // Only repaint if we need to.
    if (hover_x != old_hover_x || hover_y != old_hover_y)
        update();
}

void LifeWidget::mousePressEvent(QMouseEvent * event) {
    int x = event->x();
    x -= (x % CELL_SIZE);

    int y = event->y();
    y -= (y % CELL_SIZE);

    mouse_is_in_widget = true;

    if (template_mode) {
        x -= (template_width * CELL_SIZE) / 2;
        y -= (template_height * CELL_SIZE) / 2;
        x /= CELL_SIZE;
        y /= CELL_SIZE;

        for (unsigned int gy = 0; gy < template_height; gy++)
            for (unsigned int gx = 0; gx < template_width; gx++)
                if (x+gx < grid_width && y+gy < grid_height)
                    grid[x+gx][y+gy] = template_grid[gx][gy];
    } else {
        x /= CELL_SIZE;
        y /= CELL_SIZE;
        grid[x][y] = !grid[x][y];
    }
    hover_x = -1;
    redraw_pixmap();
    update();
}

void LifeWidget::leaveEvent(QEvent* /* event */) {
    mouse_is_in_widget = false;
    update();
}

void LifeWidget::enterEvent(QEvent* /* event */) {
    mouse_is_in_widget = true;
    update();
}

unsigned int LifeWidget::count_neighbours(unsigned int x, unsigned int y) {
    unsigned int neighbours = 0;

    if (x > 0 && y > 0 && grid[x-1][y-1]) neighbours++;
    if (y > 0 && grid[x][y-1]) neighbours++;
    if (x+1 < grid_width && y > 0 && grid[x+1][y-1]) neighbours++;

    if (x > 0 && grid[x-1][y]) neighbours++;
    if (x+1 < grid_width && grid[x+1][y]) neighbours++;

    if (x > 0 && y+1 < grid_height && grid[x-1][y+1]) neighbours++;
    if (y+1 < grid_height && grid[x][y+1]) neighbours++;
    if (x+1 < grid_width && y+1 < grid_height && grid[x+1][y+1]) neighbours++;

    return neighbours;
}

void LifeWidget::iterate(){
    grid_mutex.lock();
    copy_grid(secondary_grid, grid, grid_width, grid_height);

    for (unsigned int x = 0; x < grid_width; x++) {
        for (unsigned int y = 0; y < grid_height; y++) {
            unsigned int neighbours = count_neighbours(x,y);

            if (neighbours == 3)
                secondary_grid[x][y] = true;
            else if (neighbours < 2 || neighbours > 3)
                secondary_grid[x][y] = false;
        }
    }

    grid_t tmp_grid = grid;
    grid = secondary_grid;
    secondary_grid = tmp_grid;

    grid_mutex.unlock();

}

void LifeWidget::async_iterate() {
    QFuture<void> future = QtConcurrent::run(this,&LifeWidget::iterate);
    future.waitForFinished();
    redraw_pixmap();
    update();
}
