/*****************************************************************************
 *            Created: 10/05/2012
 *          Copyright: (C) Copyright David Goodwin, 2012
 *            License: GNU General Public License
 *****************************************************************************
 *
 *   This is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ****************************************************************************/

#include "templatebrowser.h"
#include "ui_templatebrowser.h"
#include "grid.h"

#include <QDir>
#include <QTreeWidgetItem>
#include <QStringList>
#include <QFileInfoList>
#include <QPainter>
#include <QPixmap>

#define CELL_SIZE 5

#define TEMPLATES_SUBDIR "templates"
#define TEMPLATE_EXTENSION "*.zgt"
#define NO_TEMPLATE_DESCRIPTION "Select a template on the left to place it with the template tool."
#define NO_TEMPLATE_PREVIEW "No template selected"

TemplateBrowser::TemplateBrowser(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::TemplateBrowser)
{
    ui->setupUi(this);

    connect(ui->treeWidget, SIGNAL(itemClicked(QTreeWidgetItem*,int)),
            this, SLOT(treeItemClicked(QTreeWidgetItem*,int)));

    ui->textBrowser->setPlainText(NO_TEMPLATE_DESCRIPTION);
    ui->lblPreview->setText(NO_TEMPLATE_PREVIEW);

    populateTemplates();
}

TemplateBrowser::~TemplateBrowser()
{
    delete ui;
}

void TemplateBrowser::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void TemplateBrowser::populateTemplates() {
    QDir dir(TEMPLATES_SUBDIR);

    QStringList categories = dir.entryList(QDir::AllDirs | QDir::NoDotAndDotDot);

    foreach(QString category, categories) {
        QTreeWidgetItem *category_item = new QTreeWidgetItem(
                    ui->treeWidget,
                    QStringList(category));

        ui->treeWidget->addTopLevelItem(category_item);

        QDir subdir(TEMPLATES_SUBDIR "/" + category);
        QFileInfoList templates = subdir.entryInfoList(
                    QStringList(TEMPLATE_EXTENSION),QDir::Files);

        foreach(QFileInfo template_file, templates) {
            QTreeWidgetItem* child = new QTreeWidgetItem(
                        QStringList(template_file.baseName()));
            child->setData(0,Qt::UserRole, template_file.absoluteFilePath());

            category_item->addChild(child);
        }
    }
}

void TemplateBrowser::treeItemClicked(QTreeWidgetItem *item, int /*column*/) {
    if (item == NULL)
        return;

    QVariant userData = item->data(0, Qt::UserRole);
    if (userData.isValid()) {
        ui->lblPreview->setText("");
        previewFile(userData.toString());
    } else {
        // No template selected.
        ui->textBrowser->setPlainText(NO_TEMPLATE_DESCRIPTION);
        ui->lblPreview->setText(NO_TEMPLATE_PREVIEW);
    }
}

void TemplateBrowser::previewFile(QString filename) {
    QString description;
    uint32_t width, height;
    grid_t grid = load_grid(&width, &height, filename, &description);

    currentFile = filename;

    ui->textBrowser->setPlainText(description);

    // Render a preview
    QPixmap preview = QPixmap(width * CELL_SIZE, height * CELL_SIZE);
    QPainter painter(&preview);

    painter.setPen(QPen(Qt::white));
    painter.setBrush(QBrush(Qt::white));

    painter.drawRect(0,0,preview.width(),preview.height());
    painter.setBrush(QBrush(Qt::black));

    for (unsigned int x = 0; x < width; x++)
        for (unsigned int y = 0; y < height; y++)
            if (grid[x][y])
                painter.drawRect(x*CELL_SIZE,y*CELL_SIZE,CELL_SIZE,CELL_SIZE);

    ui->lblPreview->setPixmap(preview);

    free_grid(grid, width);
    grid = 0;
    place();
}

void TemplateBrowser::place() {
    emit templateSelected(currentFile);
}
